from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from uuid import uuid4

from .. import config
from .layers import *
from .layers import AnnType, _attr_names

from openpecha.utils import dump_yaml, load_yaml

class Global2LocalId:
    """Map global id of annotation in a layer to local id of a layer."""

    def __init__(self, local_id_dict=None):
        self.start_local_id = 200_000
        self.global2local_id = self._initialize(local_id_dict)
        self.local2global_id = {
            l_id["local_id"]: g_id for g_id, l_id in self.global2local_id.items()
        }
        self.last_local_id = self.find_last()

    def _initialize(self, local_id_dict):
        g2lid = {}
        if not local_id_dict:
            return g2lid
        for global_id, local_id in local_id_dict.items():
            g2lid[global_id] = {"local_id": local_id, "is_found": False}
        return g2lid

    def find_last(self):
        """Return last local id in a layer."""
        if self.global2local_id:
            return list(self.global2local_id.values()).pop()["local_id"]
        return self.start_local_id - 1

    def add(self, global_id):
        """Map given `global_id` to the last local id."""
        next_local_id = self.last_local_id + 1
        self.global2local_id[global_id] = next_local_id
        self.last_local_id = next_local_id

    def get_local_id(self, global_id):
        """Return `local_id` associated to given `global_id`."""
        return self.global2local_id.get(global_id, None)

    def get_global_id(self, local_id):
        """Return `global_id` associated to given `local_id`."""
        global_id = self.local2global_id.get(local_id, None)
        if not global_id:
            return
        self.global2local_id[global_id]["is_found"] = True
        return global_id

    def serialize(self):
        """Return just the global and local id paris."""
        result = {}
        for global_id, id_obj in self.global2local_id.items():
            if isinstance(id_obj, int):
                result[global_id] = id_obj
            elif id_obj["is_found"]:
                result[global_id] = id_obj["is_found"]
        return result


class LocalIdManager:
    """Maintains local_id to uuid map for echa layer."""

    def __init__(self, layers):
        self.map_name = _attr_names.LOCAL_ID
        self.maps = self._get_local_id_maps(layers)

    def _get_local_id_maps(self, layers):
        maps = defaultdict(dict)
        for layer in layers:
            for vol in layers[layer]:
                maps[layer][vol] = Global2LocalId(
                    layers[layer][vol].get(self.map_name, {})
                )
        return maps

    def add(self, layer_name, vol_id, global_id):
        """Add `global_id` to layer's global2local id map."""
        if layer_name not in self.maps or vol_id not in self.maps[layer_name]:
            self.maps[layer_name][vol_id] = Global2LocalId()
        self.maps[layer_name][vol_id].add(global_id)

    def get_serialized_global2local_id(self, layer_name, vol_id):
        """Convert map of given `layer_name` in global and local id pairs."""
        serialized_dict = self.maps[layer_name][vol_id].serialize()
        del self.maps[layer_name][vol_id]
        return serialized_dict


def _name(ann_name, vols_anns):
    return [(ann_name, vol_anns) for vol_anns in vols_anns]


class BaseFormatter:
    """
    OpenPecha Base class Formatter to parse annotated text into openpecha format.

    Example of OpenPoti format
    ==========================

        W1OP000001.opf
            ├── base.txt          # plain text, without markups (annotations)
            ├── layers            # layers with annotation's char coordinate
            │   ├── title.yml
            │   ├── yigchung.yml
            |   ├── citation.yml
    """

    def __init__(self, output_path, metadata):
        self.output_path = Path(output_path if output_path else config.PECHAS_PATH)
        self.metadata = metadata

    def get_unique_id(self):
        return uuid4().hex

    def _build_dirs(self, input_path, id_=None):
        """
        Build the necessary directories for OpenPecha format.
        """
        if id_:
            self.pecha_id = id_
        else:
            self.pecha_id = input_path.stem

        self.dirs = {
            "opf_path": self.output_path / f"{self.pecha_id}/{self.pecha_id}.opf"
        }
        self.dirs["layers_path"] = self.dirs["opf_path"] / "layers"
        self.dirs["base_path"] = self.dirs["opf_path"] / "base"

        self.dirs["layers_path"].mkdir(parents=True, exist_ok=True)
        self.dirs["base_path"].mkdir(parents=True, exist_ok=True)

    @property
    def opf_path(self):
        return self.dirs["opf_path"]

    @property
    def pecha_path(self):
        return self.opf_path.parent

    @property
    def meta_fn(self):
        return self.opf_path / "meta.yml"

    def normalizeUni(self, strNFC):
        strNFC = strNFC.replace("\u0F00", "\u0F68\u0F7C\u0F7E")  # ༀ
        strNFC = strNFC.replace("\u0F43", "\u0F42\u0FB7")  # གྷ
        strNFC = strNFC.replace("\u0F48", "\u0F47\u0FB7")  # ཈
        strNFC = strNFC.replace("\u0F4D", "\u0F4C\u0FB7")  # ཌྷ
        strNFC = strNFC.replace("\u0F52", "\u0F51\u0FB7")  # དྷ
        strNFC = strNFC.replace("\u0F57", "\u0F56\u0FB7")  # བྷ
        strNFC = strNFC.replace("\u0F5C", "\u0F5B\u0FB7")  # ཛྷ
        strNFC = strNFC.replace("\u0F69", "\u0F40\u0FB5")  # ཀྵ
        strNFC = strNFC.replace("\u0F73", "\u0F71\u0F72")  # ཱི
        strNFC = strNFC.replace("\u0F75", "\u0F71\u0F74")  # ཱུ
        strNFC = strNFC.replace("\u0F76", "\u0FB2\u0F80")  # ྲྀ
        strNFC = strNFC.replace("\u0F77", "\u0FB2\u0F71\u0F80")  # ཷ
        strNFC = strNFC.replace("\u0F78", "\u0FB3\u0F80")  # ླྀ
        strNFC = strNFC.replace("\u0F79", "\u0FB3\u0F71\u0F80")  # ཹ
        strNFC = strNFC.replace("\u0F81", "\u0F71\u0F80")  # ཱྀ
        strNFC = strNFC.replace("\u0F93", "\u0F92\u0FB7")  # ྒྷ
        strNFC = strNFC.replace("\u0F9D", "\u0F9C\u0FB7")  # ྜྷ
        strNFC = strNFC.replace("\u0FA2", "\u0FA1\u0FB7")  # ྡྷ
        strNFC = strNFC.replace("\u0FA7", "\u0FA6\u0FB7")  # ྦྷ
        strNFC = strNFC.replace("\u0FAC", "\u0FAB\u0FB7")  # ྫྷ
        strNFC = strNFC.replace("\u0FB9", "\u0F90\u0FB5")  # ྐྵ
        return strNFC

    def text_preprocess(self, text):
        raise NotImplementedError(
            "Text preprocessing depends on type of text format, \
                                   should be implemented in sub-class."
        )

    def get_input(self, input_path):
        """
        Return a preprocess text from given input_file path
        """
        m_text = self.text_preprocess(input_path.read_text())
        return self.normalizeUni(m_text)

    def get_old_layers(self, new_layers):
        layers = defaultdict(dict)
        for layer in new_layers:
            for vol in self.dirs["layers_path"].iterdir():
                vol_layer_fn = vol / f"{layer}.yml"
                if not vol_layer_fn.is_file():
                    continue
                layers[layer][vol.name] = load_yaml(vol_layer_fn)
        return layers

    def _inc_layer_revision(self, layer):
        inc_rev_int = int(layer["revision"]) + 1
        layer["revision"] = f"{inc_rev_int:05}"

    def add_new_ann(self, layer, vol_id, ann):
        uuid = self.get_unique_id()
        layer["annotations"][uuid] = ann
        self.local_id_manager.add(layer["annotation_type"], vol_id, uuid)

    def _add_local_id(self, layer, vol_id):
        layer[
            _attr_names.LOCAL_ID
        ] = self.local_id_manager.get_serialized_global2local_id(
            layer[_attr_names.ANNOTATION_TYPE], vol_id
        )

    def create_new_layer(self, layer_name, vol_id, anns):
        new_layer = Layer(self.get_unique_id(), layer_name)
        for _, ann in anns:
            self.add_new_ann(new_layer, vol_id, ann)
        self._add_local_id(new_layer, vol_id)
        return new_layer

    def _remove_deleted_anns(self, layer, vol_id):
        global2local_id = self.local_id_manager.maps[
            layer[_attr_names.ANNOTATION_TYPE]
        ][vol_id].global2local_id
        for global_id, id_obj in global2local_id.items():
            if isinstance(id_obj, int) or id_obj["is_found"]:
                continue
            else:
                del layer[_attr_names.ANNOTATION][global_id]

    def update_layer(self, layer, anns, vol_id):
        self._inc_layer_revision(layer)
        for local_id, ann in anns:
            if local_id:
                uuid = self.local_id_manager.maps[layer["annotation_type"]][
                    vol_id
                ].get_global_id(local_id)
                if uuid:
                    for key, value in ann.items():
                        layer["annotations"][uuid][key] = value
            # Local_id missing, possible cases
            # 1. New Annotation created
            # 2. Local_id gets deleted
            else:
                self.add_new_ann(layer, vol_id, ann)
                # TODO: implement case 2

        self._remove_deleted_anns(layer, vol_id)

    def _get_vol_layers(self, layers):
        for layer_name in layers:
            if layer_name in [AnnType.topic, AnnType.sub_topic]:
                continue
            layers[layer_name] = _name(layer_name, layers[layer_name])
        return zip(*layers.values())

    def format_layer(self, layers):
        old_layers = self.get_old_layers(layers)
        self.local_id_manager = LocalIdManager(old_layers)

        # filter cross vols layers from layers
        cross_vols_layers = {}
        for cross_ann_name in [AnnType.topic, AnnType.sub_topic]:
            cross_vols_layers[cross_ann_name] = layers[cross_ann_name]
            del layers[cross_ann_name]

        # Create Annotaion layers
        for (i, vol_layers) in enumerate(self._get_vol_layers(layers)):
            vol_id = f"v{i+1:03}"
            result = {}
            for layer_name, vol_layer_anns in vol_layers:
                if not vol_layer_anns:
                    continue
                if vol_id in old_layers[layer_name]:
                    vol_old_layer = old_layers[layer_name][vol_id]
                    self.update_layer(vol_old_layer, vol_layer_anns, vol_id)
                    result[layer_name] = vol_old_layer
                else:
                    result[layer_name] = self.create_new_layer(
                        layer_name, vol_id, vol_layer_anns
                    )

            yield result, vol_id

        if AnnType.topic not in old_layers:
            # Create Index layer
            Index_layer = Layer(self.get_unique_id(), "index")
            # loop over each topic
            for topics, sub_topics in zip(
                cross_vols_layers[AnnType.topic], cross_vols_layers[AnnType.sub_topic]
            ):
                if topics:
                    Topic = deepcopy(Text)
                    Topic["parts"] += sum(
                        [[ann for none_local_id, ann in anns] for anns in sub_topics],
                        [],
                    )
                    Topic["span"] += [ann["span"] for none_local_id, ann in topics]
                    Topic["work_id"] = topics[0][1]["work_id"]
                    uuid = self.get_unique_id()
                    Index_layer["annotations"][uuid] = Topic

            yield {"index": Index_layer}, None
        else:
            yield None, None

    def build_layers(self, text):
        """
        Parse all the layers annotation from the given text.
        """
        raise NotImplementedError(
            "Parsing annotation depends type of annotation in the text, \
                                  should be implemented in sub-class."
        )

    def get_base_text(self, m_text):
        "Retuns text with all annotation removed"
        raise NotImplementedError(
            "Every type of text have different format for annotation, \
                                  should be implemented in sub_class."
        )

    def create_opf(self, input_path):
        input_path = Path(input_path)
        self._build_dirs(input_path)

        m_text = self.get_input(input_path)
        layers = self.build_layers(m_text)
        base_text = self.get_base_text(m_text)

        # save layers
        for layer, ann in layers.items():
            layer_fn = self.dirs["layers_path"] / f"{layer}.yml"
            dump_yaml(ann, layer_fn)

        # save base_text
        (self.dirs["opf_path"] / "base.txt").write_text(base_text)



import rdflib
import requests
import pyewts
import csv
from rdflib import BNode, Literal, URIRef
from rdflib.namespace import OWL, RDF, RDFS, SKOS, XSD, Namespace, NamespaceManager
import boto3
import botocore
import gzip
import hashlib
import io
import json
import logging

LDSPDIBASEURL = "https://ldspdi.bdrc.io/"
CONVERTER = pyewts.pyewts()

SESSION = boto3.Session()
S3 = SESSION.client('s3')

BDR = Namespace("http://purl.bdrc.io/resource/")
BDO = Namespace("http://purl.bdrc.io/ontology/core/")
BDA = Namespace("http://purl.bdrc.io/admindata/")
ADM = Namespace("http://purl.bdrc.io/ontology/admin/")

def fetch_op_commits(ldspdibaseurl="http://ldspdi.bdrc.io/"):
    """
    Fetches the list of all openpecha commits on BUDA
    """
    res = {}
    headers = {"Accept": "text/csv"}
    params = {"format": "csv"}
    with closing(
        requests.get(
            ldspdibaseurl + "/query/table/OP_allCommits",
            stream=True,
            headers=headers,
            params=params,
        )
    ) as r:
        reader = csv.reader(codecs.iterdecode(r.iter_lines(), "utf-8"))
        for row in reader:
            if not row[0].startswith("http://purl.bdrc.io/resource/IE0OP"):
                logging.error("cannot interpret csv line starting with " + row[0])
                continue
            res[row[0][34:]] = row[1]
    return res


def get_s3_folder_prefix(wlname, image_group_lname):
    """
    gives the s3 prefix (~folder) in which the volume will be present.
    inpire from https://github.com/buda-base/buda-iiif-presentation/blob/master/src/main/java/
    io/bdrc/iiif/presentation/ImageInfoListService.java#L73
    Example:
       - wlname=W22084, image_group_lname=I0886
       - result = "Works/60/W22084/images/W22084-0886/
    where:
       - 60 is the first two characters of the md5 of the string W22084
       - 0886 is:
          * the image group ID without the initial "I" if the image group ID is in the form I\\d\\d\\d\\d
          * or else the full image group ID (incuding the "I")
    """
    md5 = hashlib.md5(str.encode(wlname))
    two = md5.hexdigest()[:2]

    pre, rest = image_group_lname[0], image_group_lname[1:]
    if pre == 'I' and rest.isdigit() and len(rest) == 4:
        suffix = rest
    else:
        suffix = image_group_lname

    return 'Works/{two}/{RID}/images/{RID}-{suffix}/'.format(two=two, RID=wlname, suffix=suffix)

def gets3blob(s3Key):
    f = io.BytesIO()
    try:
        S3.download_fileobj('archive.tbrc.org', s3Key, f)
        return f
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == '404':
            return None
        else:
            raise

def get_image_list_s3(wlname, image_group_lname):
    s3key = get_s3_folder_prefix(wlname, image_group_lname)+"dimensions.json"
    blob = gets3blob(s3key)
    if blob is None:
        return None
    blob.seek(0)
    b = blob.read()
    ub = gzip.decompress(b)
    s = ub.decode('utf8')
    data = json.loads(s)
    return data

def get_image_list_iiifpres(wlname, image_group_lname):
    r = requests.get(f"http://iiifpres.bdrc.io/il/v:bdr:{vol_name}")
    return r.json()

def get_image_list(wlname, image_group_lname, source="s3", reorder_with_bvm=False):
    il = None
    if source == "s3":
        il = get_image_list_s3(wlname, image_group_lname)
    else:
        il = get_image_list_iiifpres(wlname, image_group_lname)
    return il

def _res_from_model(g, wlname):
    res = {
        "source_metadata": {
            "id": "http://purl.bdrc.io/resource/"+wlname
        },
        "image_groups": {}
    }
    wres = BDR[wlname]
    try:
        adm = g.value(predicate=ADM.adminAbout, object=wres)
        res["source_metadata"]["status"] = str(g.value(adm, ADM.status))
        res["source_metadata"]["access"] = str(g.value(adm, ADM.access))
        if (adm, ADM.restrictedInChina, Literal(True)) in g:
            res["source_metadata"]["geo_restriction"] = ["CN"]
        mwres = g.value(wres, BDO.instanceReproductionOf)
        res["source_metadata"]["reproduction_of"] = str(mwres)
        for _, _, cs in g.triples((mwres, BDO.copyright, None)):
            res["source_metadata"]["copyright_status"] = str(cs)
        if "copyright_status" not in res["source_metadata"]:
            res["source_metadata"]["copyright_status"] = "http://purl.bdrc.io/resource/CopyrightPublicDomain"
        res["source_metadata"]["reproduction_of"] = str(mwres)
        for _, _, l in g.triples((mwres, SKOS.prefLabel, None)):
            if l.language == "bo-x-ewts":
                res["source_metadata"]["title"] = CONVERTER.toUnicode(l.value)
                break
            else:
                res["source_metadata"]["title"] = l.value
        res["source_metadata"]["languages"] = set()
        for _, _, wa in g.triples((mwres, BDO.instanceOf, None)):
            for _, _, l in g.triples((wa, BDO.language, None)):
                for _, _, lt in g.triples((l, BDO.langBCP47Lang, None)):
                    res["source_metadata"]["languages"].add(lt.value)
            for _, _, aac in g.triples((wa, BDO.creator, None)):
                if (aac, BDO.role, BDR.R0ER0009) or (aac, BDO.role, BDR.R0ER0009) in g:
                    for _, _, p in g.triples((aac, BDO.agent, None)):
                        for _, _, l in g.triples((p, SKOS.prefLabel, None)):
                            if l.language == "bo-x-ewts":
                                res["source_metadata"]["author"] = CONVERTER.toUnicode(l.value)
                                break
                            else:
                                res["source_metadata"]["author"] = l.value
        res["source_metadata"]["languages"] = list(res["source_metadata"]["languages"])
        for _, _, ig in g.triples((wres, BDO.instanceHasVolume, None)):
            iglname = str(ig)[str(ig).rfind('/')+1:]
            res["image_groups"][iglname] = {}
            iginfo = res["image_groups"][iglname]
            iginfo["id"] = str(ig)
            iginfo["total_pages"] = int(g.value(ig, BDO.volumePagesTotal))
            iginfo["volume_number"] = int(g.value(ig, BDO.volumeNumber))
            iginfo["volume_pages_bdrc_intro"] = int(g.value(ig, BDO.volumePagesTbrcIntro))
            for _, _, l in g.triples((ig, SKOS.prefLabel, None)):
                if l.language == "bo-x-ewts":
                    iginfo["title"] = CONVERTER.toUnicode(l.value)
                    break
                else:
                    iginfo["title"] = l.value
    finally:
        return res

def get_buda_scan_info(wlname):
    headers = {"Accept": "text/turtle"}
    params = {"R_RES": "bdr:"+wlname}
    res = None
    g = rdflib.Graph()
    try:
        req = requests.get(
            LDSPDIBASEURL + "query/graph/OP_info",
            headers=headers,
            params=params,
        )
        g.parse(data=req.text, format="ttl")
        res = _res_from_model(g, wlname)
    except Exception as e:
        logging.error("get_buda_scan_info failed for "+wlname+": "+str(e))
    finally:
        return res

def image_group_to_folder_name(scan_id, image_group_id):
    image_group_folder_part = image_group_id
    pre, rest = image_group_id[0], image_group_id[1:]
    if pre == "I" and rest.isdigit() and len(rest) == 4:
        image_group_folder_part = rest
    return scan_id+"-"+image_group_folder_part


"""OpenPecha Work

This module contains OpenPecha Work data model and utils for
creating work objects.
"""
import json
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Union

from pydantic import BaseModel, validator
from pydantic.networks import AnyHttpUrl

from .. import config
from ..core.ids import get_work_id
from ..core.metadata import InitialCreationType
from ..utils import download_pecha, dump_yaml, load_yaml


class InstanceMetadata(BaseModel):
    """OpenPecha Work instance metadata."""

    source_id: str
    initial_creation_type: InitialCreationType
    title: str
    author: str = None
    organization: AnyHttpUrl
    created_at: datetime
    last_modified_at: datetime


class Instance(BaseModel):
    """Instance model

    This model represents a single instance of a work.
    """

    id: str
    metadata: InstanceMetadata


class WorkNotFound(Exception):
    """Work not found exception."""

    pass


class Work(BaseModel):
    """OpenPecha Work

    This class contains OpenPecha Work data model.
    """

    id: str = None
    bdrc_id: str = None
    title: str
    alternative_titles: List[str] = []
    author: str = None
    isRoot: bool = None
    language: str = None
    type_defination: str = None
    wikidata_id: str = None
    instances: Dict[str, Dict] = {}

    @validator("id", pre=True, always=True)
    def set_id(cls, v):
        return v or get_work_id()

    def add_instance(self, instance: Instance):
        """Add instance to work.

        Args:
            instance (Dict): Instance to add to work.
        """
        self.instances[instance.id] = instance.metadata.dict()

    def get_instance(self, instance_id: str) -> Instance:
        """Get instance from work.

        Args:
            instance_id (str): Instance id to get from work.

        Returns:
            instance: Instance from work.
        """
        metadata = self.instances.get(instance_id)
        if not metadata:
            return None
        return Instance(id=instance_id, metadata=InstanceMetadata(**metadata))

    def remove_instance(self, instance_id: str):
        """Remove instance from work.

        Args:
            instance_id (str): Instance id to remove from work.
        """
        del self.instances[instance_id]

    @classmethod
    def from_yaml(cls, fn: Union[str, Path]) -> "Work":
        """Create Work from yaml file.

        Args:
            fn (str): YAML file name.

        Returns:
            Work: Work object.
        """
        fn = Path(fn)
        data = load_yaml(fn)
        work = cls.parse_obj(data)
        for instance_id, metadata in work.instances.items():
            work.instances[instance_id] = InstanceMetadata.parse_obj(metadata)
        return work

    def save_to_yaml(self, output_dir: Union[str, Path] = None) -> Path:
        """Save work to yaml file.

        Args:
            output_dir (str): Output directory.

        Returns:
            Path: Path to saved yaml file.
        """
        if not output_dir:
            output_dir = download_pecha(config.WORKS_REPO_NAME)
        output_dir = Path(output_dir)
        work_fn = output_dir / f"{self.id}.yaml"
        dump_yaml(data=json.loads(self.json()), output_fn=work_fn)
        return work_fn

    @classmethod
    def from_id(cls, id_: str) -> "Work":
        """Create Work from id.

        Args:
            id (str): Work id.

        Returns:
            Work: Work object.
        """
        works_path = download_pecha(config.WORKS_REPO_NAME)
        work_fn = works_path / f"{id_}.yaml"
        if not work_fn.is_file():
            work_fn = works_path / f"{id_}.yml"
            if not work_fn.is_file():
                raise WorkNotFound(f"Work {id_} not found")
        return cls.from_yaml(work_fn)



"""
This module contains format variable for all the annotations
"""

from collections import namedtuple

__all__ = [
    "Layer",
    "Page",
    "Span",
    "Text",
    "Correction",
    "ErrorCandidate",
    "CrossVolSpan",
    "SubText",
    "Peydurma",
    "Chapter",
    "Tsawa",
    "Citation",
    "Sabche",
    "Yigchung",
    "BookTitle",
    "SubTitle",
    "BookNumber",
    "PotiTitle",
    "Author",
    "Archaic",
    "Span",
    "Footnote",
]


class AnnType:
    book_title = "BookTitle"
    sub_title = "SubTitle"
    book_number = "BookNumber"
    poti_title = "PotiTitle"
    author = "Author"
    chapter = "Chapter"

    topic = "Text"
    sub_topic = "SubText"

    pagination = "Pagination"
    citation = "Citation"
    correction = "Correction"
    error_candidate = "ErrorCandidate"
    peydurma = "Peydurma"
    pedurma_note = "PedurmaNote"
    sabche = "Sabche"
    tsawa = "Tsawa"
    yigchung = "Yigchung"
    archaic = "Archaic"
    durchen = "Durchen"
    footnote = "Footnote"


class _attr_names:
    # Layer
    ID = "id"  # Uique id for annotation of specific Pecha or Abstract work. type: str
    ANNOTATION_TYPE = "annotation_type"  # Name of annotation, type: str
    REVISION = "revision"  # Revision number. type: int
    ANNOTATION = "annotations"  # Annotations are stored in list . type: dict
    LOCAL_ID = "local_ids"

    # Spans
    SPAN = "span"
    VOL = "vol"
    START = "start"
    END = "end"

    # Page
    IMGNUM = "imgnum"
    PAGE_INDEX = "page_index"  # Page number based on Volume specified, type: int
    PAGE_REFERENCE = "reference"  # Any reference of page, eg: img_url. type: str

    # Text
    WORK_ID = "work_id"

    # Correction
    CORRECTION = "correction"  # Correct text suggestion. type: str
    CERTAINTY = "certainty"  # Certainty of the suggested correct text. type: int

    # Peydurma
    NOTE = "note"  # syls, word or phrase to be compared to other publication

    # Archaic word
    MODERN = "modern"

    # Tsawa, Citation
    ISVERSE = "isverse"  # Boolean flag to indicate a sache in verse format or not

    # Footnote
    FOOTNOTE_REF = "footnote_ref"


def Layer(id_, type_, rev=f"{1:05}"):
    return {
        _attr_names.ID: id_,
        _attr_names.ANNOTATION_TYPE: type_,
        _attr_names.REVISION: rev,
        _attr_names.ANNOTATION: {},
    }


def Span(start, end):
    return {_attr_names.START: start, _attr_names.END: end}


# ~~~ INDEX Layer ~~~~~
# Cross vol Span
CrossVolSpan = {"vol": None, "span": None}


def CrossVolSpan(vol, start, end):
    return {_attr_names.VOL: vol, _attr_names.START: start, _attr_names.END: end}


# Sub_text annotation
SubText = {"work": None, "span": []}  # index of the sub_text  # span of the sub_text

# Text annotation
Text = {
    "parts": [],
    "span": [],
    "work_id": "",
}  # list of SubText  # list of CrossVolSpan

# ~~~~~ ANNOTATIONS Layers ~~~~~~


def Page(span, imgnum=None, page_ref=None):
    return {
        _attr_names.IMGNUM: imgnum,
        _attr_names.PAGE_REFERENCE: page_ref,
        _attr_names.SPAN: span,
    }


def Correction(span, correction=None, certainty=None):
    return {
        _attr_names.CORRECTION: correction,
        # _attr_names.CERTAINTY: certainty,
        _attr_names.SPAN: span,
    }


def Archaic(span, modern=None, certainty=None):
    return {
        _attr_names.MODERN: modern,
        # _attr_names.CERTAINTY: certainty,
        _attr_names.SPAN: span,
    }


def ErrorCandidate(span):
    return {_attr_names.SPAN: span}


def Peydurma(span, note=None):
    return {_attr_names.NOTE: note, _attr_names.SPAN: span}


def BookTitle(span):
    return {_attr_names.SPAN: span}


def SubTitle(span):
    return {_attr_names.SPAN: span}


def BookNumber(span):
    return {_attr_names.SPAN: span}


def PotiTitle(span):
    return {_attr_names.SPAN: span}


def Author(span):
    return {_attr_names.SPAN: span}


def Chapter(span):
    return {_attr_names.SPAN: span}


def Tsawa(span, isverse=False):
    return {_attr_names.SPAN: span, _attr_names.ISVERSE: isverse}


def Citation(span, isverse=False):
    return {_attr_names.SPAN: span, _attr_names.ISVERSE: isverse}


def Sabche(span, isverse=False):
    return {_attr_names.SPAN: span}


def Yigchung(span):
    return {_attr_names.SPAN: span}


def Footnote(span, footnote_ref):
    return {_attr_names.SPAN: span, _attr_names.FOOTNOTE_REF: footnote_ref}
